@@ -3,15 +3,14 @@ package Mojolicious::Plugin::AccessLog;
 use Mojo::Base 'Mojolicious::Plugin';
 use Mojo::IOLoop;
 
-use Carp qw(carp croak);
 use File::Spec;
 use IO::File;
 use POSIX qw(setlocale strftime LC_ALL);
-use Scalar::Util qw(blessed reftype);
+use Scalar::Util qw(blessed reftype weaken);
 use Socket qw(inet_aton AF_INET);
 use Time::HiRes qw(gettimeofday tv_interval);
 
-our $VERSION = '0.006';
+our $VERSION = '0.007';
 
 my $DEFAULT_FORMAT = 'common';
 my %FORMATS = (
@@ -31,6 +30,7 @@ my $TZOFFSET = strftime('%z', localtime) !~ /^[+-]\d{4}$/ && do {
 sub register {
     my ($self, $app, $conf) = @_;
     my $log = $conf->{log} // $app->log->handle;
+    my ($pkg, $f, $l) = caller 2;   # :-/
     my $fh;
 
     unless ($log) { # somebody cleared $app->log->handle?
@@ -58,7 +58,8 @@ sub register {
             or $log = $app->home->rel_file($log);
 
         $fh = IO::File->new($log, '>>')
-            or croak qq{Can't open log file "$log": $!};
+            or die <<"";
+Can't open log file "$log": $! at $f line $l.
 
         $fh->autoflush(1);
         $logger = sub { $fh->print($_[0]) };
@@ -69,7 +70,9 @@ sub register {
     }
 
     if ($conf->{uname_helper}) {
-        carp 'uname_helper is DEPRECATED in favor of $c->req->env->{REMOTE_USER}';
+        warn <<"";
+uname_helper is DEPRECATED in favor of \$c->req->env->{REMOTE_USER} at $f line $l.
+
 
         my $helper_name = $conf->{uname_helper};
 
@@ -104,7 +107,8 @@ sub register {
     }
 
     # each handler is called with following parameters:
-    # ($tx, $tx->req, $tx->res, $tx->req->url, $time, $bytes_in, $bytes_out)
+    # 0: $tx, 1: $tx->req, 2: $tx->res, 3: $tx->req->url,
+    # 4: $request_start_time, 5: $process_time, 6: $bytes_in, 7: $bytes_out
 
     my $block_handler = sub {
         my ($block, $type) = @_;
@@ -115,7 +119,19 @@ sub register {
         return sub { $_[2]->headers->header($block) // '-' }
             if $type eq 'o';
 
-        return sub { '[' . $strftime->($block, localtime) . ']' }
+        return sub {
+            return $_[4][0]
+                if $block eq 'sec';
+            return sprintf "%u%03u", $_[4][0], int($_[4][1] / 1000)
+                if $block eq 'msec';
+            return sprintf "%u%06u", @{$_[4]}
+                if $block eq 'usec';
+            return sprintf('%03u', $_[4][1] / 1000)
+                if $block eq 'msec_frac';
+            return sprintf('%06u', $_[4][1])
+                if $block eq 'usec_frac';
+            return $strftime->($block, localtime($_[4][0]));
+        }
             if $type eq 't';
 
         return sub { _safe($_[1]->cookie($block // '')) }
@@ -136,34 +152,30 @@ sub register {
         a => $remoteaddr_cb,
         A => sub { $_[0]->local_address // '-' },
         b => sub {
-            $_[2]->headers->content_length ||
-            $_[6] - $_[2]->header_size - $_[2]->start_line_size ||
-            '-'
+            $_[7] && ($_[7] - $_[2]->header_size - $_[2]->start_line_size) || '-'
         },
         B => sub {
-            $_[2]->headers->content_length ||
-            $_[6] - $_[2]->header_size - $_[2]->start_line_size
+            $_[7] ? $_[7] - $_[2]->header_size - $_[2]->start_line_size : '0'
         },
-        D => sub { int($_[4] * 1000000) },
+        D => sub { int($_[5] * 1000000) },
         h => $remoteaddr_cb,
         H => sub { 'HTTP/' . $_[1]->version },
-        I => sub { $_[5] },
+        I => sub { $_[6] },
         l => '-',
         m => sub { $_[1]->method },
-        O => sub { $_[6] },
+        O => sub { $_[7] },
         p => sub { $_[0]->local_port },
         P => sub { $$ },
         q => sub {
             my $s = $_[3]->query->to_string or return '';
             return '?' . $s;
         },
-        r => sub {
-            $_[1]->method . ' ' . _safe($_[3]->to_string) .
-            ' HTTP/' . $_[1]->version
+        r => sub { substr($_[1]->build_start_line, 0, -2) },
+        s => sub { $_[2]->code // '-' },
+        t => sub {
+            $strftime->('[%d/%b/%Y:%H:%M:%S %z]', localtime($_[4][0]))
         },
-        s => sub { $_[2]->code },
-        t => sub { '[' . $strftime->('%d/%b/%Y:%H:%M:%S %z', localtime) . ']' },
-        T => sub { int $_[4] },
+        T => sub { int $_[5] },
         u => sub {
             my $env = $_[1]->env;
             my $user =
@@ -186,13 +198,10 @@ sub register {
         };
     }
 
-    my ($time_stats);
     my $char_handler = sub {
         my $char = shift;
         my $cb = $char_handler{$char};
 
-        $time_stats = 1 if $char eq 'T' or $char eq 'D';
-
         return $char_handler{$char} if $char_handler{$char};
 
         $app->log->error("\%$char not supported.");
@@ -215,26 +224,28 @@ sub register {
 
     $app->hook(after_build_tx => sub {
         my $tx = $_[0];
-        my $t; $t = [gettimeofday] if $time_stats;
         my $bcr = my $bcw = 0;
-        my ($s, $r, $w);
+        my ($r, $s, $t, $w);
 
         $tx->on(connection => sub {
             my ($tx, $connection) = @_;
 
+            $t = [gettimeofday];
             $s = Mojo::IOLoop->stream($connection);
             $r = $s->on(read  => sub { $bcr += length $_[1] });
             $w = $s->on(write => sub { $bcw += length $_[1] });
         });
 
+        # watch for the right moment to fetch the un-expanded start-line
+        $tx->req->once(progress => sub { $_[0]->build_start_line });
+
         $tx->on(finish => sub {
             my $tx = shift;
-
-            $t = tv_interval($t) if $time_stats;
+            my $dt = tv_interval($t);
 
             $s->unsubscribe(read  => $r);
             $s->unsubscribe(write => $w);
-            $logger->(_log($tx, $format, \@handler, $t, $bcr, $bcw));
+            $logger->(_log($tx, $format, \@handler, $t, $dt, $bcr, $bcw));
         });
     });
 }
@@ -267,7 +278,7 @@ Mojolicious::Plugin::AccessLog - AccessLog Plugin
 
 =head1 VERSION
 
-Version 0.006
+Version 0.007
 
 =head1 SYNOPSIS
 
@@ -424,7 +435,7 @@ Time the request was received (standard english format).
 
 =item %T
 
-Custom field for handling times in subclasses.
+The time taken to serve the request, in seconds.
 
 =item %u
 
@@ -464,8 +475,39 @@ The contents of response header C<ResponseHeaderName>.
 
 =item %{Format}t
 
-The time, in the form given by C<Format>, which should be in
-L<strftime(3)> format.
+The time, in the form given by C<Format>, which should be in extended
+L<strftime(3)> format (potentially localized). In addition to the
+formats supported by strftime(3), the following format tokens are
+supported:
+
+=over
+
+=item sec:
+
+Number of seconds since the Epoch.
+
+=item msec:
+
+Number of milliseconds since the Epoch.
+
+=item usec:
+
+Number of microseconds since the Epoch.
+
+=item msec_frac:
+
+Millisecond fraction.
+
+=item usec_frac:
+
+Microsecond fraction.
+
+=back
+
+These tokens can not be combined with each other or strftime(3) formatting
+in the same format string. You can use multiple %{format}t tokens instead:
+
+  "%{%d/%b/%Y %T}t.%{msec_frac}t %{%z}t"
 
 =item %{CookieName}C
 
